/*
 * Copyright (c) 2009 Jess Nuevo Chiquero - UAH
 *
 * This file is part of RawVideoTools
 *
 * RawVideoTools is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * RawVideoTools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <errno.h>

#include <fstream>

#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>

#include <dirent.h>

#include <iostream>

#include <cv.h>
#include <cxcore.h>
#include <highgui.h>

#include <time.h>


#include "frame_header.hh"
#include "raw_video_reader.hh"

static float diff_time_ms(timeval32 &t2, timeval32 &t1)
{
    return (double)(t2.tv_sec - t1.tv_sec)*1.0e3 + (double)(t2.tv_usec - t1.tv_usec)/1.0e3;
}


int RawVideoReader::open(const char *fn)
{

    int version = 1;

    if(m_video){
	fclose(m_video);
	reset();
   }

    if(!fn){
	m_ok = false;
	return 1;
    }

    m_video = fopen(fn, "r");
    if(!m_video){
	std::cerr << "No se pudo abrir el archivo RAW. "<< strerror(errno) << std::endl;
	reset();
	return -1;
    }
    m_fn = fn;

    version = read_frame_header_version(m_video);

    if(read_frame_header(m_video, &m_head)){
	std::cerr << "Video RAW erroneo: BAD START_CODE"
		  << std::endl;
	reset();
	return -1;
    }

    if(m_head.ncameras==0)
	m_head.ncameras++;

    if(version==3 && m_head.ncameras > 4) //bug
	m_head.ncameras = 1;

    CvSize sz = cvSize(m_head.width, m_head.ncameras* m_head.height);
    int nchannels;


    if(!strncmp(m_head.color_space, "GRAY", 4))
	nchannels = 1;
    else if(!strncmp(m_head.color_space, "RGB", 3) || !strncmp(m_head.color_space, "BGR",3))
	nchannels = 3;

    else if(m_head.version>2){//check if the color_space is "raw8"
	if(!strncmp(m_head.color_space, "raw8",4))
	    nchannels = 1;
    }
    else{
	std::cerr << "Unknown color space "<<m_head.color_space<<std::endl;
	return -1;
    }

    m_img = cvCreateImage(sz, IPL_DEPTH_8U, nchannels);
    
    {
	fseeko(m_video,0,SEEK_END);
	m_length = ftello(m_video)/(long long)m_img->imageSize;	
	fseeko(m_video,0, SEEK_SET);
    }

//     m_position = 0;
    m_frame = 0;
    m_ok = true;

    estimate_fps();

    return 0;
      
}

const IplImage * RawVideoReader::query_frame()
{

    if(!m_ok || !m_video){
	std::cerr << "No video" << std::endl;
	return NULL;
    }

    m_prev_time = m_head.filltime;
 
    size_t r = fread((void*)m_img->imageData, sizeof(char), m_img->imageSize, m_video);

     
    if(!r || feof(m_video)){
	std::cout << "end of video" << std::endl;
	return NULL;
    }
    if(r!=(uint)m_img->imageSize){
	std::cerr << "Error reading raw video: "<< r << "read, expected: "
		      << m_img->imageSize << std::endl;
	if(feof(m_video)){
	    std::cerr << "Unspected end of video" << std::endl;
	}
	m_ok = false;
	return NULL;
    }

    m_frame = ftello(m_video)/(long long)m_img->imageSize;
    
    check_frame_header(m_img, m_frame, &m_head);    
 
    return m_img;
}


double RawVideoReader::get_property(int property_id)
{

    if(!m_ok || !m_video){
	std::cerr << "No video" << std::endl;
	return -1;
    }

    switch( property_id )
	{
	case CV_CAP_PROP_POS_MSEC:
	    return 1e3*(double)m_frame/m_avg_fps;
	    break;
	case CV_CAP_PROP_FRAME_COUNT:
	    return m_length;
	    break;
	case CV_CAP_PROP_POS_FRAMES:
	    return m_frame;
	    break;
	case CV_CAP_PROP_POS_AVI_RATIO:
	    return (double)m_frame/(double)m_length;
	    break;
	case CV_CAP_PROP_FRAME_WIDTH:
	    return m_img->width;
	    break;
	case CV_CAP_PROP_FRAME_HEIGHT:
	    return m_img->height;
	    break;
	case CV_CAP_PROP_FPS:
	    return m_avg_fps;
	    break;
	case CV_CAP_PROP_FOURCC:
	    return (double)CV_FOURCC('Y', '8', '0', '0');
	    break;
	}
    return 0;
    
}


int RawVideoReader::set_property(int property_id, double value)
{

    if(!m_ok || !m_video){
	std::cerr << "No video" << std::endl;
	return -1;
    }


    switch( property_id )
	{
	case CV_CAP_PROP_POS_MSEC:
	case CV_CAP_PROP_POS_FRAMES:
	case CV_CAP_PROP_POS_AVI_RATIO:
	    {
		long n= m_frame;
				
		switch( property_id )
		    {
		    case CV_CAP_PROP_POS_MSEC:			
			n = (long)floor(value*m_avg_fps / 1e3);
			break;
		    case CV_CAP_PROP_POS_FRAMES:
			n = (long)floor(value);
			break;
		    case CV_CAP_PROP_POS_AVI_RATIO:
			if(value>=0 && value <1)
			    n = (long)floor(value*m_length);			
			break;			
		    }
		//skip the video to the new position
		if(n>=0 && (unsigned long long)n< m_length && (unsigned long long)n!=m_frame){
		    off_t pos;
		    pos = n * (long long)m_img->imageSize;
		    fseeko(m_video, pos, SEEK_SET);

		    //comprobar que la cabecera est ok...

		    m_frame = n;
		}
		//if the requested position is past the end, place the pointer at the end
		else if (n <=0)
		    fseeko(m_video, 0, SEEK_SET);
		else if((unsigned long)n >= m_length){
		    fseeko(m_video, 0, SEEK_END);		    
		}
	    }
	}
    return 0;

}


void RawVideoReader::estimate_fps()
{

    off_t init;
    off_t end;
    
    frame_header  check;
    timeval32 pt;
    int err;

    double fps;
    uint ai;

    init = ftello(m_video);
    fseeko(m_video, 0, SEEK_END);
    end = ftello(m_video);
    fseeko(m_video, init, SEEK_SET);

    m_avg_fps = 0;
    
    //read 10 headers and get the mean time.

    for(ai=0; ai<10 && ai < end && ftello(m_video)<end; ai++){
	pt = check.filltime;
	err = read_frame_header(m_video, &check);
	
	fps = 1e3/diff_time_ms(check.filltime, pt);

	if(ai>0){
	    m_avg_fps = (m_avg_fps*(ai-1) + fps)/ai;
	}
	
	fseeko(m_video, m_img->imageSize, SEEK_CUR);
    }

    fseeko(m_video, init, SEEK_SET);


    //check for bad values

    if(m_avg_fps<1e-3 || m_avg_fps>1e3)
	m_avg_fps = 30;

    return;
}
