/*
 * Copyright (c) 2009 Jess Nuevo Chiquero - UAH
 *
 * This file is part of RawVideoTools
 *
 * RawVideoTools is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * RawVideoTools is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef RAW_VIDEO_READER_HH
#define RAW_VIDEO_READER_HH


#include <string.h>

#include <stdlib.h>
#include <stdio.h>
#include <error.h>
#include <errno.h>

#include <fstream>

#include <sys/time.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>

#include <dirent.h>


#include <cv.h>
#include <cxcore.h>
#include <highgui.h>

#include <time.h>
#include "frame_header.hh"


/**
 * \class RawVideoReader
 *
 * A class to read mono RAW videos, recorded with the Robesafe header format (by Pedro Jimenez)
 * 
 */

class RawVideoReader{

public:
    RawVideoReader(){m_img = NULL; reset();}

    /** 
     * Open a video
     * 
     * @param fn The path to the video
     * 
     * @return 0 on success, or an error otherwise: 1 if fn is NULL,
     * or -1 if the file can not be open or an invalid header is found
     */
    int open(const char *fn);
    
    void close(){ 
	if(m_video)fclose(m_video); 
	reset(); }

    /** 
     * Check the file is open and accesible.
     * 
     * 
     * @return true or false
     */
    bool good(){return m_ok && m_video;}

    /** 
     * Extract a frame from the video
     * 
     * @return The image. It should not be modified by the user.
     */
    const IplImage * query_frame();


    /** 
     * Get a property of the video stream.
     *
     * @param property The property identifier. Use the same values as
     * for OpenCV's cvGetCaptureProperty, see the documentation of
     * HighGUI. The frame rate is calculated as the difference between
     * the current frame and the previous one.
     *
     * 
     * @return 
     */
    double get_property(int property);

    /** 
     * Set a property of the video stream.
     * 
     * @param property The property to be set. Only properties related
     * to the stream position are available. (\c CV_CAP_PROP_POS_MSEC,
     * \c CV_CAP_PROP_POS_FRAMES, \c CV_CAP_PROP_POS_AVI_RATIO)
     *
     * @param value The value of the property.
     * 
     * @return 
     */
    int set_property(int property, double value);
    
    ~RawVideoReader(){if(m_video)fclose(m_video); reset();}

protected:
    
    unsigned long m_frame; //current frame number
    unsigned long m_length; //length in frames

    double m_avg_fps; //average frame rate

    FILE *m_video;
    std::string m_fn;

    IplImage *m_img; //buffer image

    bool m_ok;

    frame_header m_head;
    timeval32 m_prev_time;

    void reset() {
	m_frame = 0;
	m_length = 0;
	m_avg_fps = -1;
	m_video = NULL;	
	m_fn.clear();
	m_ok = false;	
	if(CV_IS_IMAGE(m_img))
	    cvReleaseImage(&m_img);
    }

    void estimate_fps();

};





#endif
